import numpy as np
import os
from ..stream_builder import StreamBuilder
from ..stream import Stream, intertwine_dwells


def get_spot_dwells(t, position, max_dwt=5):
    """Gets a single spot dwell

    Args:
        t (float): Time in seconds for the spot dwell.
        position ((2,) list, array): Position of the dwell.
        max_dwt (float, optional): Maximum allowed dwell time in ms. Defaults to 5.

    Returns:
        (n,3) array: Dwells
    """
    dwells_split = StreamBuilder.split_dwells(
        np.array([1000 * t, position[0], position[1]])[np.newaxis, :], max_dwt)
    return np.vstack(dwells_split)


def get_spot_calibration(grid=[5, 3], start_time=1, end_time=15, shuffle=True, addressable_pixels=[65536, 56576], max_dwt=5):
    """Gets the spot calibration stream

    Args:
        grid (list, optional): Grid size for distributing the structures. Defaults to [5, 3].
        start_time (int, optional): Time in seconds of the smallest structure. Defaults to 1.
        end_time (int, optional): Time in seconds of the largest structure. Defaults to 15.
        shuffle (bool, optional): Whether to shuffle the structures randomly on a screen. Defaults to True.
        addressable_pixels (list, optional): Addressable pixels on microscope screen. Defaults to [65536, 56576].
        max_dwt (float, optional): Maximum allowed dwell time in ms. Defaults to 5.

    Returns:
        tuple: Stream, data
    """
    n_structs = grid[0] * grid[1]
    times = np.linspace(start_time, end_time, n_structs)
    # randomly shuffle the times to get rid of transient effects
    if shuffle:
        np.random.shuffle(times)
    # get the positions of the streams
    xstep, ystep = np.array(addressable_pixels) * 0.8 / np.array(grid)
    x = np.arange(grid[0]) * xstep
    y = np.arange(grid[1]) * ystep
    xygrid = np.meshgrid(x, y)
    positions = np.array(xygrid).reshape(2, n_structs)

    # construct the streams
    spot_dwells = [get_spot_dwells(t, positions[:, i], max_dwt=max_dwt)
                   for i, t in enumerate(times)]
    # intertwine the stream dwells
    dwells = intertwine_dwells(spot_dwells)
    # combine
    combined_stream = Stream(dwells, addressable_pixels=addressable_pixels,
                             max_dwt=max_dwt)
    combined_stream.recentre()

    # save the dwell times and the positions
    positions_loc = np.array(np.meshgrid(
        np.arange(grid[0]), np.arange(grid[1]))).reshape(2, n_structs).T
    data = np.hstack((times[:, np.newaxis], positions_loc))
    return combined_stream, data


def export_spot_calibration(file_path, **kwargs):
    """Convinience function to immediately save the calibration stream. Takes all the keywords parameter of the get_spot_calibration

    Args:
        file_path (str): File to which to save.
    """
    strm, data = get_spot_calibration(**kwargs)
    strm.print_time()
    strm.write(file_path)
    data_path = os.path.splitext(file_path)[0] + '_data.csv'
    np.savetxt(data_path, data, delimiter='\t',
               header="dwellTime\tpositionX\tpositionY", comments='', fmt="%3f\t%d\t%d")




from mpl_toolkits import mplot3d
import numpy as np
import trimesh
from .plotting import create_3d_axes, points3d, set_axes_equal
from .slicing import split_eqd
from .branches import split_intersection, get_branch_connections
import numpy as np
from trimesh.intersections import mesh_multiplane
import time
from scipy.spatial.transform import Rotation


class Structure(trimesh.Trimesh):
    """Class defining the mesh structure. Inherits from trimesh.base.Trimesh class.


        Attributes:
            file_path (str): Path to the STL file.
            pitch (float): Pitch for the slicing
            fill (bool): If true attempts to fill in the STL file (not implemented yet)
    """

    def __init__(self, *args, file_path=None, pitch=3, fill=False, **kwargs) -> None:
        if not 'face_colors' in kwargs:
            kwargs['face_colors'] = (17, 103, 177)
        super().__init__(*args, **kwargs)
        self.pitch = pitch
        self.fill = fill

        self.file_path = file_path
        self.clear_slicing()

    @classmethod
    def from_file(cls, file_path: str, **kwargs):
        """Gets the structure from stl file.

        Args:
            file_path (str): Path to the stl file.
            **kwargs: Further settings (pitch, fill) are passed to the __init__ method of the class.

        Returns:
            Structure
        """
        file_path = file_path
        msh = trimesh.load_mesh(file_path, file_type='stl')
        # create the structure and add file path
        struct = cls(vertices=msh.vertices, faces=msh.faces,
                     file_path=file_path, **kwargs)
        return struct

    @property
    def slices(self):
        """List of arrays of (n,2) points in each slice 
        """
        if self._slices is None:
            self.generate_slices()
        return self._slices

    @property
    def branches(self):
        """List of arrays signifying to which branch does each point in slice correspond to.
        """
        if self._branches is None:
            self.generate_slices()
        return self._branches

    @property
    def branch_lengths(self):
        """Branch lengths.
        """
        if self._branch_lengths is None:
            self.generate_slices()
        return self._branch_lengths

    @property
    def branch_connections(self):
        """Branch connection. List of lists. branch_connections[i][j] is the list of indices of which branches in the slice i-1 is the branch j in slice i connected.
        """
        if self._branch_connections is None:
            self.generate_slices(branch_connectivity=True)
        return self._branch_connections

    @property
    def z_levels(self):
        """Array of z values where the slices are.
        """
        if self._z_levels is None:
            self.generate_slices()
        return self._z_levels

    @property
    def dz_slices(self):
        """The thickness of layers
        """
        return self.z_levels[1:] - self.z_levels[:-1]

    @property
    def is_sliced(self) -> bool:
        """Weather or not the structure is sliced.

        Returns:
            bool
        """
        if self.slices is None:
            return False
        return True

    def centre(self):
        """Centres the structure to (0, 0)
        """
        self.rezero()
        transf = np.zeros((4, 4)).astype(float)
        transf[:3, :3] = np.eye(3)
        transf[0, -1] = -self.centroid[0]
        transf[1, -1] = -self.centroid[1]
        self.apply_transform(transf)

    def rescale(self, scale):
        """Scales the mesh by the scale factor"""
        transf = np.eye(4)
        transf[3, 3] = 0
        transf *= scale
        self.apply_transform(transf)
        self.clear_slicing()

    def rotate(self, rotation_axis, rotation_angle):
        """
        Rotates the mesh by a given angle around a specified axis.

        Args:
            rotation_axis: list or np.array
                Specifies rotation axis [x, y, z] as a 3-vector.
            rotation_angle: float
                Rotation angle in degrees.
        """
        r = Rotation.from_rotvec(np.deg2rad(
            rotation_angle) * np.array(rotation_axis))
        transf_matrix = np.eye(4)
        transf_matrix[:3, :3] = r.as_matrix()
        self.apply_transform(transf_matrix)
        self.clear_slicing()

    def clear_slicing(self):
        """Clears the slicing of the structure.
        """
        self._slices = None
        self._branches = None
        self._branch_lengths = None
        self._branch_connections = None
        self._z_levels = None

    def plot_mpl(self, ax=None):
        """Plots the mesh vertices in matplotlib window.

        Returns:
            axes: Matplotlib axes.
        """
        if ax is None:
            ax = create_3d_axes()

        ax.add_collection3d(mplot3d.art3d.Poly3DCollection(self.triangles))
        bounds = self.bounds

        ax.set_xlim(bounds[0, 0], bounds[1, 0])
        ax.set_ylim(bounds[0, 1], bounds[1, 1])
        ax.set_zlim(bounds[0, 2], bounds[1, 2])
        set_axes_equal(ax)
        return ax

    def get_3dslices(self):
        """Gets the slices with appended z values

        Returns:
            slices (list of (n,3) arrays)
        """
        return [np.hstack(
            [sl, z * np.ones((sl.shape[0], 1))]) for sl, z in zip(self.slices, self.z_levels)]

    def get_sliced_points(self):
        """Gets the sliced points in a matrix form.

        Returns:
            points ((n, 3) array)
        """
        points = np.vstack(self.get_3dslices())
        return points

    def plot_slices(self, *args, **kwargs):
        """Plots the slices in matplotlib.

        Returns:
            axes: Matplotlib axes.
        """
        points = self.get_sliced_points()
        ax = points3d(points, *args, **kwargs)
        return ax

    def show(self):
        # shows the model and adds the build plate
        scene = self.scene()
        # add the plane to the scene
        length = np.max(self.extents[:2])
        plane = trimesh.creation.box([length, length, 1])
        scene.add_geometry(plane)
        scene.set_camera(angles=np.deg2rad([45, 0, 0]))
        return scene.show()

    def generate_slices(self, branch_connectivity=True):
        """Gets the silces and all the corresponding information.

        Args:
            branch_connectivity (bool, optional): If false, does not calculate
            the connectivity of branches required for resistance calculations.
            This can be useful to save time if resistance is not going to be calculated.
            Defaults to True.
        """
        print('Slicing...')
        intersection_lines, self._z_levels = self.get_intersection_lines()

        # split into connected components (branches)
        branch_intersections_slices = [split_intersection(
            inter) for inter in intersection_lines]

        # get branch connectivity. This is the slowest part and might not be necessary if not doing the resistance.
        # t0 = time.time()
        if branch_connectivity:
            connection_distance = self.pitch + 0.01
            self._branch_connections = get_branch_connections(
                branch_intersections_slices, connection_distance)
        # print(time.time() - t0)

        # get equally separated points and branch indices
        self._slices, self._branches, self._branch_lengths = split_eqd(
            branch_intersections_slices, self.pitch)

        print('Sliced')

    def get_intersection_lines(self):
        """Gets the intersections and z_levels.

        Returns:
            intersection_lines (list of arrays): A list of (n,2,2) arrays representing the intersection lines as start_node-end_node
            z_levels (array): Array of z levels corresponding to the intersections.
        """
        mindz = self.pitch / 5
        maxdz = 2 * self.pitch
        slice_height = self.pitch / 2

        # get the heights of slices
        minz, maxz = self.bounds[0, 2], self.bounds[1, 2]
        # move a bit to avoid artifacts
        minz += 1e-3
        maxz -= 1e-3
        z_levels = np.arange(minz, maxz, slice_height)

        # define the slicing plane
        plane_normal = np.array((0.0, 0.0, 1.0))
        plane_orig = np.zeros(3).astype(float)

        intersection_lines, _, _ = mesh_multiplane(
            self, plane_orig, plane_normal, z_levels)
        # drop the empty intersections
        nonempty = np.array(
            len(inter) != 0 for inter in intersection_lines).astype(bool)
        z_levels = z_levels[nonempty].flatten()
        intersection_lines = [
            inter for inter in intersection_lines if len(inter) != 0]
        return intersection_lines, z_levels



import numpy as np
from .utils import load_settings
import io
from datetime import timedelta
import os
from .plotting import points2d

# conversion factor from ms to 0.1us
CONVERSION_FACTOR = 10000


def intertwine_dwells(dwells_list):
    """Takes the list of matrices of dwells and intertwines them.

    Args:
        dwells_list (list of (n,3 arrays)): Dwells to intertwine.

    Returns:
        (n,3) array: Array of intertwined dwells.
    """
    n_list = len(dwells_list)
    size_list = [dwls.shape[0] for dwls in dwells_list]
    max_rows = np.max(size_list)
    total_length = np.sum(size_list)
    dwells = np.zeros((total_length, 3))

    cnt = 0
    for i in range(max_rows):
        for j in range(n_list):
            if i >= dwells_list[j].shape[0]:
                continue
            dwells[cnt, :] = dwells_list[j][i, :]
            cnt += 1
    return dwells


class Stream:
    """Class representing the stream file.

        Attributes:
            dwells ((n,3) array): Specifying dwells (t, x, y) in (ms, px, px)
            addressable_pixels (list of two int): Microscope addressable pixels.
            max_dwt (float): Maximum dwell time in ms.
    """

    def __init__(self, dwells, addressable_pixels=[65536, 56576], max_dwt=5):
        self.dwells = dwells

        self.addressable_pixels = addressable_pixels
        self.max_dwt = max_dwt

    @classmethod
    def from_file(cls, file_path, **kwargs):
        """Imports the stream from .str file

        Args:
            file_path (str): Path to the .str file

        Returns:
            Stream: Stream class with dwells from the file.
        """
        dwells = np.genfromtxt(file_path, delimiter=' ',
                               skip_header=3, usecols=range(3))
        # convert the dwells to ms
        dwells /= CONVERSION_FACTOR
        return cls(dwells, **kwargs)

    @property
    def limits(self):
        """Limits of the stream in x and y directions.

        Returns:
            (2, 2) array: Limits of the stream
        """
        mnmx = np.zeros((2, 2))
        mnmx[:, 0] = np.min(self.dwells[:, 1:], axis=0)
        mnmx[:, 1] = np.max(self.dwells[:, 1:], axis=0)
        return mnmx

    def is_valid(self):
        """Checks if the stream is valid (i.e. within the bounds).

        Returns:
            bool:
        """
        limits = self.limits
        if np.any(limits[:, 0] < 0) or np.any(limits[:, 1] > self.addressable_pixels) or np.any(self.dwells[:, 0] > self.max_dwt):
            return False
        return True

    def recentre(self, position=None):
        """Centres the stream.

        Args:
            position ((2,) array, optional): Position on which to centre. If None, centres to the centre of the screen. Defaults to None.
        """
        stream_centre = (self.limits[:, 1] + self.limits[:, 0]) / 2
        screen_centre = np.array(self.addressable_pixels) / 2
        if position is None:
            position = screen_centre
        else:
            position = np.array(position)
        translation_vector = position - stream_centre
        self.dwells[:, 1:] += translation_vector[np.newaxis, :]

    def write(self, file_path, centre=True):
        """Writes the stream to the file_path.

        Args:
            file_path (str): Path to the .str file to which to write.
            centre (bool, optional): Whether to centre the stream before exporting. Defaults to True.
        """
        # make sure the extension is .str
        file_path = os.path.splitext(file_path)[0] + '.str'
        if centre:
            self.recentre()
        # check that all the points are within limits
        if not self.is_valid():
            raise Exception(
                'Stream not valid! One of the dimensions is out of range.')

        dwells_to_write = self.dwells.copy()
        dwells_to_write[:, 0] *= CONVERSION_FACTOR
        dwells_to_write = np.round(dwells_to_write).astype(int)

        # gets the string ready to be written. Slightly roudabout way, but it's because of optional blanked screen lines
        header = 's16\n1\n' + str(self.dwells.shape[0])
        bio = io.BytesIO()
        np.savetxt(bio, dwells_to_write.astype(int),
                   delimiter=' ', fmt='%d', header=header, comments='')
        dwls_string = bio.getvalue().decode('latin1')
        dwls_string = dwls_string[:-1] + ' 0'
        with open(file_path, 'w') as f:
            f.write(dwls_string)

    def show_on_screen(self):
        """Plots the stream as it would look on the microscope screen.
        """
        ax, sc = points2d(self.dwells[:, 1:])
        ax.set_xlabel('x [px]')
        ax.set_ylabel('y [px]')
        ax.set_xlim([0, self.addressable_pixels[0]])
        ax.set_ylim([0, self.addressable_pixels[1]])

    def get_time(self):
        """Gets the total stream time.

        Returns:
            datetime.timedelta:
        """
        return timedelta(milliseconds=np.sum(self.dwells[:, 0]))

    def print_time(self):
        """Prints the total stream time."""
        print('Total time: ', str(self.get_time()))



# MOVE DOCS1 TO DOCS BOTTOM OF THIS FILE


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import inspect


def abspath(rel):
    """
    Take paths relative to the current file and
    convert them to absolute paths.

    Parameters
    ------------
    rel : str
      Relative path, IE '../stuff'

    Returns
    -------------
    abspath : str
      Absolute path, IE '/home/user/stuff'
    """

    # current working directory
    cwd = os.path.dirname(os.path.abspath(
        inspect.getfile(inspect.currentframe())))
    return os.path.abspath(os.path.join(cwd, rel))


extensions = ['sphinx.ext.napoleon', 'nbsphinx']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates', 'templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = ['.rst']

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'F3AST'
copyright = '2021, Luka Skoric'
author = 'Luka Skoric'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

# grab from f3ast without installing
exec(open(abspath('../f3ast/version.py')).read())
version = __version__
# The full version, including alpha/beta/rc tags.
release = __version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# -- Options for HTML output --------------------------------------

# The theme to use for HTML and HTML Help pages
html_theme = 'sphinx_rtd_theme'

# options for rtd-theme
html_theme_options = {
    'display_version': True,
    'prev_next_buttons_location': 'bottom',
    'style_external_links': False,
    # toc options
    'collapse_navigation': True,
    'sticky_navigation': True,
    'navigation_depth': 4,
    'includehidden': True,
    'titles_only': False,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Output file base name for HTML help builder.
htmlhelp_basename = 'f3astdoc'

# -- Extensions configuration ----------------------------------------

autodoc_default_options = {
    'autosummary': True,
}



import pytest

from f3ast import Structure, load_settings, DDModel, StreamBuilder, Stream


@pytest.fixture
def settings():
    return load_settings()


@pytest.fixture
def structure_file():
    return "test/simple_ramp.stl"


@pytest.fixture
def structure(structure_file, settings):
    return Structure.from_file(
        structure_file, **settings["structure"])


def test_structure_slicing(structure):
    structure.generate_slices()
    assert structure.is_sliced


@pytest.fixture
def model(structure, settings):
    gr = 0.15
    k = 1
    sigma = 4.4
    return DDModel(structure, gr, k, sigma, **settings['dd_model'])


def test_stream(model, settings):
    stream_builder, dwell_solver = StreamBuilder.from_model(
        model, **settings['stream_builder'])
    strm = stream_builder.get_stream()
    assert isinstance(strm, Stream)



import numpy as np
import numpy.linalg as la


def get_resistance(struct, single_pixel_width=50.):
    """Gets the resistance for each slice and for each point in a slice.

    Args:
        struct (Structure)
        single_pixel_width (float, optional): Width of a single pixel line. Defaults to 50.

    Returns:
        list of (n,) arrays: Resistance per point in the slice.
    """

    slices = struct.slices
    branches = struct.branches
    branch_lengths = struct.branch_lengths
    branch_connections = struct.branch_connections
    dz_levels = struct.z_levels[1:] - struct.z_levels[:-1]
    resistance_slices = []
    for i, (pts, branch, brlens, conn) in enumerate(zip(slices, branches, branch_lengths, branch_connections)):
        resistance = np.zeros(pts.shape[0])
        unique_br = np.unique(branch)
        dz = dz_levels[i - 1]
        # separate the points into branches
        separated_pts = [pts[branch == lbl, :] for lbl in unique_br]
        # if on the substrate, set the resistance to 0 and continue
        if i == 0:
            separated_pts_below = separated_pts
            resistances_below = resistance
            resistance_slices.append(resistance)
            continue

        # find out how the branches connect: for each of the branches find the
        # ones from the layer below it connects to by finding if there are any
        # points which are within the pitch. Then add their resistance in
        # parallel to the other branches it connects to
        for j, (br_pts, br_conn) in enumerate(zip(separated_pts, conn)):
            # get the centre of the current layer
            center_curr = np.mean(br_pts, axis=0)
            # for each of the connected layers below, get the resistance increment
            r_inv = 0
            for c in br_conn:
                branch_pts_below = separated_pts_below[c]
                center_below = np.mean(branch_pts_below, axis=0)
                # get the layer separation (in um)
                layer_sep = np.sqrt(
                    la.norm(center_curr - center_below)**2 + dz**2) / 1000
                # get the connection resistance, normalize by the single pixel width
                connection_resistance = resistances_below[c] + \
                    single_pixel_width * layer_sep / \
                    (brlens[j] + single_pixel_width)
                r_inv += 1 / connection_resistance
            r = 1 / r_inv if r_inv != 0 else 0
            resistance[branch == j] = r
        resistance_slices.append(resistance)
        separated_pts_below = separated_pts
        resistances_below = resistance
    return resistance_slices



# This is a set of functions for splitting slices into connected components and determining how these components connect with each other
from scipy.spatial import KDTree
from joblib.parallel import Parallel, delayed
import numpy as np
from trimesh.grouping import group_rows
from trimesh.graph import connected_component_labels


def split_intersection(intersection):
    """Splits the intersection into connected components (branches).

    Args:
        intersection ((n,2,2)): array of intersection lines

    Returns:
        (m,) list of (k,2,2) arrays: intersections grouped into components
    """
    # group the same points in the intersections
    grouped_rows = group_rows(intersection.reshape(-1, 2))
    # assign each point an index.
    grouped_indices = np.array(
        [[l, i] for i, ls in enumerate(grouped_rows) for l in ls])
    # get the indices sorted so that i-th element of node_indices corresponds to the i-th point
    try:
        arg = np.argsort(grouped_indices[:, 0])
    except IndexError:
        print(len(intersection))
        print(intersection)
        raise Exception()
    node_indices = grouped_indices[arg, 1]
    # label the connected components
    edges = node_indices.reshape(-1, 2)
    conn_labels = connected_component_labels(
        edges, node_count=len(grouped_rows))
    # conn labels correspond to the nodes. Label each edge by one of its nodes.
    edge_labels = conn_labels[edges[:, 0]]

    unique_lbls = np.unique(conn_labels)
    split_intersection = []
    for lbl in unique_lbls:
        split_intersection.append(intersection[edge_labels == lbl, :, :])
    return split_intersection


def get_branch_connections(branch_intersections_slices, connection_distance):
    """Gets the connections between branches organized in slices

    Args:
        branch_intersections_slices (list of lists of arrays): For each slice, for each branch, array of points in that branch.
        connection_distance (float): Distance for which branches are considered connected.

    Returns:
        list of list of arrays:  For each slice, for each branch, which branches from layer below is it connected to.
    """
    branch_connections = []
    # with Parallel(n_jobs=5, backend="threading") as parallel:
    for i, separated_pts in enumerate(branch_intersections_slices):
        if i == 0:
            branch_connections.append([])
            separated_pts_below = separated_pts
            continue
        this_slice_connections = []
        for br_pts in separated_pts:
            # get how the branches in this layer connect to the branches from the previous layer. This is done by checking if the branches are neighbours (within a connection_distance away)
            this_branch_connections = []
            tree = KDTree(br_pts.reshape(-1, 2))
            branch_min_distances = []
            for k, branch_pts_below in enumerate(separated_pts_below):
                nb, dist = is_branch_nb(
                    tree, branch_pts_below, connection_distance)
                if nb:
                    branch_min_distances.append(dist)
                    this_branch_connections.append(k)
            this_branch_connections = np.array(this_branch_connections)
            n_conn = len(this_branch_connections)

            # the previous has an issue. The two branches can be close, but not fully merge until a few layers above.
            # This is a hack to fix it. It looks if the branches that seem to connect to each other exist (within a connection distance) in the current layer.
            # If something does not work properly, this is the likely culprit.
            # There must be a better way of doing this
            if n_conn > 1:
                # Ensure that the branches that were merged do not exist in the current layer
                # to know how many duplicates, need to know how many branches in this layer are within a conn distance away
                count = 0
                for br_pts2 in separated_pts:
                    if is_branch_nb(tree, br_pts2, connection_distance)[0]:
                        count += 1
                if count > 1 and count <= n_conn:
                    # drop the extra connections
                    n_drop = n_conn - count + 1
                    keep_indx = np.argsort(branch_min_distances)[:n_drop]
                    this_branch_connections = this_branch_connections[keep_indx]
            this_slice_connections.append(this_branch_connections)
        branch_connections.append(this_slice_connections)
        separated_pts_below = separated_pts
    return branch_connections


def is_branch_nb(tree, branch_pts, connection_distance):
    """Gets if the branch is a neighbour of a branch with KDTree tree.

    Args:
        tree (KDTree): KDTree of the current branch
        branch_pts ((n, 2) array): Other branch pts
        connection_distance (float): Distance for which the branches are considered neighours

    Returns:
        bool: True if branches are close
        float: Minimal distance between branches. np.inf if not neighbours.
    """
    tree1 = KDTree(branch_pts.reshape(-1, 2))
    distance_matrix = tree.sparse_distance_matrix(
        tree1, connection_distance, p=np.inf)
    if distance_matrix.count_nonzero() > 0:
        min_dist = np.min(list(distance_matrix.values()))
        return True, min_dist
    return False, np.inf


def get_branch_connections_in_slice(br_pts, separated_pts_below, connection_distance):
    """Auxiliary function to get the connections in a slice. This is for the purposes of parallelizing, but is not used atm.
    """

    this_branch_connections = []
    tree = KDTree(br_pts.reshape(-1, 2))
    branch_min_distances = []
    for k, branch_pts_below in enumerate(separated_pts_below):
        nb, dist = is_branch_nb(
            tree, branch_pts_below, connection_distance)
        if nb:
            branch_min_distances.append(dist)
            this_branch_connections.append(k)
    this_branch_connections = np.array(this_branch_connections)
    return np.array(this_branch_connections), branch_min_distances



import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
import numpy as np
import numpy.linalg as la


def create_3d_axes():
    """Creates 3D axes.

    Returns:
        axes: 3D axes.
    """
    fig = plt.figure()
    ax = mplot3d.Axes3D(fig, auto_add_to_figure=False)
    fig.add_axes(ax)
    return ax


def points3d(verts, *args, ax=None, equal_axes=True, **kwargs):
    """Plots the 3D points in a 3D scatter plot.

    Args:
        verts ((n,3) array): Points to plot
        ax (axes, optional): Axes on which to plot. If None creates them. Defaults to None.

    Returns:
        tuple: axes, scatter_plot
    """
    if ax is None:
        fig = plt.figure()
        ax = mplot3d.Axes3D(fig, auto_add_to_figure=False)
        fig.add_axes(ax)
    sc = ax.scatter(verts[:, 0], verts[:, 1], verts[:, 2], *args, **kwargs)
    if equal_axes:
        set_axes_equal(ax)
    return ax, sc


def plot_dwells(dwells):
    """Plots the dwells as 3D points. Colours them by time.

    Args:
        dwells ((n,4) array): Dwells to plot.

    Returns:
        tuple: axes, scatter_plot
    """
    ax, sc = points3d(dwells[:, 1:], c=dwells[:, 0], cmap='magma')
    plt.colorbar(sc, ax=ax, shrink=0.6, label='t [ms]')
    set_axes_equal(ax)
    return ax, sc


def points2d(verts, ax=None):
    """Plots the 2D points in a 2D scatter plot.

    Args:
        verts ((n,2) array): Points to plot
        ax (axes, optional): Axes on which to plot. If None creates them. Defaults to None.

    Returns:
        tuple: axes, scatter_plot
    """
    if ax is None:
        fig, ax = plt.subplots()
    sc = ax.scatter(verts[:, 0], verts[:, 1])
    return ax, sc


def set_axes_equal(ax):
    """Make axes of 3D plot have equal scale so that spheres appear as spheres,
    cubes as cubes, etc..  This is one possible solution to Matplotlib's
    ax.set_aspect('equal') and ax.axis('equal') not working for 3D.

    Args:
        ax (axes):
    """

    x_limits = ax.get_xlim3d()
    y_limits = ax.get_ylim3d()
    z_limits = ax.get_zlim3d()

    x_range = abs(x_limits[1] - x_limits[0])
    x_middle = np.mean(x_limits)
    y_range = abs(y_limits[1] - y_limits[0])
    y_middle = np.mean(y_limits)
    z_range = abs(z_limits[1] - z_limits[0])
    z_middle = np.mean(z_limits)

    # The plot bounding box is a sphere in the sense of the infinity
    # norm, hence I call half the max range the plot radius.
    plot_radius = 0.5 * max([x_range, y_range, z_range])

    ax.set_xlim3d([x_middle - plot_radius, x_middle + plot_radius])
    ax.set_ylim3d([y_middle - plot_radius, y_middle + plot_radius])
    ax.set_zlim3d([z_middle - plot_radius, z_middle + plot_radius])



from numpy.core.numeric import full
from .stream import Stream
from .solver import DwellSolver
import numpy as np
import warnings


class StreamBuilder:
    """Builds the stream using the microscope settings.


        Attributes:
            dwells_slices (list of (n,3) arrays): Specifying per layer dwells (t, x, y)
            addressable_pixels (list of two int): Microscope addressable pixels.
            max_dwt (float): Maximum dwell time in ms.
            cutoff_time (float): Minimum dwell time in ms. This is just for cutting of insignificant dwells to reduce file size.
            screen_width (float): Screen width in nm.
            scanning_order (str): Layer scanning order. Can be "serpentine" or "serial".
    """

    def __init__(self, dwells_slices,
                 addressable_pixels=[65536, 56576],
                 max_dwt=5,
                 cutoff_time=0.01,
                 screen_width=6400,
                 scanning_order="serpentine"):
        self.dwells_slices = dwells_slices

        self.addressable_pixels = addressable_pixels
        self.max_dwt = max_dwt
        self.cutoff_time = cutoff_time
        self.screen_width = screen_width
        assert scanning_order in {
            "serial", "serpentine"}, "Unrecognized scanning order!"
        self.scanning_order = scanning_order

    @classmethod
    def from_model(cls, model, **kwargs):
        """Creates the class from the model. Internally creates the DwellSolver and solves for dwells.

        Args:
            model (Model): Class defining the growth model.

        Returns:
            tuple:
                stream_builder (StreamBuilder), dwell_solver (DwellSolver)
        """
        # get the dwells
        dwell_solver = DwellSolver(model)
        dwell_solver.solve_dwells()
        dwells_slices = dwell_solver.get_dwells_slices()
        # build the class
        stream_builder = cls(dwells_slices, **kwargs)
        return stream_builder, dwell_solver

    @ property
    def ppn(self):
        """Pixels per nanometer"""
        return self.addressable_pixels[0] / self.screen_width

    def get_stream(self, centre=False):
        """Builds the stream object from the calculated dwells

        Args:
            centre (bool, optional): Wether to centre the stream on the screen. Defaults to False.

        Returns:
            Stream:
        """
        dwells = self.get_stream_dwells()
        # if the dwells include the z direction, get rid of that
        if dwells.shape[1] > 3:
            dwells = dwells[:, :3]
        stream = Stream(
            dwells, addressable_pixels=self.addressable_pixels, max_dwt=self.max_dwt)
        if centre:
            stream.recentre()
            if not stream.is_valid():
                warnings.warn(
                    "Stream outside screen limits. Structure might be too large!")
        return stream

    def get_stream_dwells(self):
        """Gets the stream dwells by splitting and ordering them appropriately. Also converts x, y in pixels and gets rid of small dwells.

        Returns:
            (n,3) array: Array of dwells.
        """
        # remove the dwells that are below the cutoff time
        dwells_slices = [ds[ds[:, 0] > self.cutoff_time]
                         for ds in self.dwells_slices]
        # split the dwells
        split_dwells_slices = [self.split_dwells(
            ds, self.max_dwt) for ds in dwells_slices]
        del dwells_slices

        # connect the split slices in a list, reverse the order of every other one if the serpentine order is used
        if self.scanning_order == "serial":
            full_dwells_list = [
                dwls for sds in split_dwells_slices for dwls in sds]
        else:
            full_dwells_list = []
            i = 0
            for sds in split_dwells_slices:
                for dwls in sds:
                    if i % 2 == 1:
                        full_dwells_list.append(np.flipud(dwls))
                    else:
                        full_dwells_list.append(dwls)
        # concatenate the list
        stream_dwells = np.vstack(full_dwells_list)
        # convert nm to px
        stream_dwells[:, 1:] *= self.ppn
        return stream_dwells

    @staticmethod
    def split_dwells(dwells, max_dwt):
        """Takes a matrix of dwells and splits them so that none of them exceeds the max dwell time. Returns a list of N_reps items which are all the split dwells.

        Args:
            dwells ((n,3) array): Array of dwells
            max_dwt (float): Maximum allowed dwell time.

        Returns:
            list: List of equal (n,3) arrays that when summed correspond to the dwells.
        """
        n_splits = int(np.ceil(np.max(dwells[:, 0]) / max_dwt))
        dwells_reduced = dwells.copy()
        dwells_reduced[:, 0] = dwells_reduced[:, 0] / n_splits
        return [dwells_reduced for i in range(n_splits)]



from f3ast.utils import load_settings
import numpy as np
from ..structure import Structure
from scipy.spatial.transform import Rotation as R
from ..stream_builder import StreamBuilder
from ..stream import Stream
import os

dirname = os.path.dirname(__file__)
CUBE_PATH = os.path.join(dirname, 'cube.stl')


def get_sigma_structures(model, sigma_list, width=75, length=800, angle=45):
    """Gets the structures for sigma calibration and returns a single stream file

    Args:
        model : Deposit model
        sigma_list (list): List of sigma values to try
        width (float, optional): Width of the structures. Defaults to 75.
        length (float, optional): Length of the structures. Defaults to 800.
        angle (float, optional): Angle to xy plane of the structures. Defaults to 45.
    """
    settings = load_settings()

    # get the straight ramp of minimal thickness
    struct = get_straight_ramp(length, width, 0.1, angle)
    model.set_structure(struct)

    # solve for dwell times
    sigma_strm_list = []
    for s in sigma_list:
        model.sigma = s
        stream_builder, _ = StreamBuilder.from_model(
            model, **settings['stream_builder'])
        sigma_strm_list.append(stream_builder.get_stream())

    # get the single pixel line
    struct_1px = get_straight_ramp(length, 0.1, 0.1, 45)
    model.set_structure(struct_1px)
    stream_builder, _ = StreamBuilder.from_model(
        model, **settings['stream_builder'])
    strm_1px = stream_builder.get_stream()

    # arange on a screen
    addressable_pixels = settings['stream_builder']['addressable_pixels']
    y_positions = np.linspace(
        0.1 * addressable_pixels[1], 0.9 * addressable_pixels[1], len(sigma_list))
    positions_list = [(addressable_pixels[0] / 2, y) for y in y_positions]
    for strm, pos in zip(sigma_strm_list, positions_list):
        strm.recentre(position=pos)

    pos1px = [0.75 * addressable_pixels[0], addressable_pixels[1] / 2]
    strm_1px.recentre(position=pos1px)

    # combine into one screen
    all_dwells = np.vstack([strm.dwells for strm in sigma_strm_list])
    all_dwells = np.vstack((all_dwells, strm_1px.dwells))

    strm_out = Stream(all_dwells, addressable_pixels=addressable_pixels,
                      max_dwt=settings['stream_builder']['max_dwt'])
    return strm_out


def get_straight_ramp(length, width, thickness, angle):
    """Gets a straight ramp stl file.

    Args:
        length (float): Length of the ramp.
        width (float): Width of the ramp.
        thickness (float): Thickness of the ramp.
        angle (float): Angle to substrate.

    Returns:
        Structure: The ramp with desired parameters.
    """
    struct = Structure.from_file(CUBE_PATH)
    transf_matrix = np.eye(4)
    transf_matrix[3, 3] = 0  # don't translate
    transf_matrix[0, 0] = length
    transf_matrix[1, 1] = width
    transf_matrix[2, 2] = thickness
    r = R.from_rotvec(np.deg2rad(angle) * np.array([0, 1, 0]))
    transf_matrix[:3, :3] = r.as_matrix().dot(transf_matrix[:3, :3])

    struct.apply_transform(transf_matrix)
    return struct



from .structure import Structure
from .deposit_model import *
from .stream import Stream
from .stream_builder import StreamBuilder
from .solver import DwellSolver
from .utils import *
from .plotting import *
from .calibration import *

from types import ModuleType
import sys
import importlib


def deep_reload(m: ModuleType):
    name = m.__name__  # get the name that is used in sys.modules
    name_ext = name + '.'  # support finding sub modules or packages

    def compare(loaded: str):
        return (loaded == name) or loaded.startswith(name_ext)

    # prevent changing iterable while iterating over it
    all_mods = tuple(sys.modules)
    sub_mods = filter(compare, all_mods)
    for pkg in sorted(sub_mods, key=lambda item: item.count('.'), reverse=True):
        # reload packages, beginning with the most deeply nested
        importlib.reload(sys.modules[pkg])



import hjson
import pickle
import os
from glob import glob
from warnings import warn
from shutil import copy2
from pathlib import Path

# TODO: remove os and just use pathlib


def load_settings(file_path='settings.hjson'):
    """Loads the settings from the given path.

    Args:
        file_path (str, optional): Path to hjson file containing the settings. Defaults to 'settings.hjson'.

    Returns:
        dict: settings dictionary
    """
    settings_path = Path(file_path)
    if not settings_path.exists():
        warn("Settings file not found. Creating a default one.")
        default_settings_path = Path(
            __file__).parent.resolve() / "default_settings.hjson"
        copy2(default_settings_path, settings_path)

    with open(file_path, 'r') as f:
        settings = hjson.load(f)
    return settings


def save_build(file_path, dwell_solver, stream_builder):
    """Saves the dwell_solver and stream_builder to the save path.

    Args:
        file_path (str): Path to which to save
        dwell_solver (DwellSolver)
        stream_builder (StreamBuilder)
    """
    assert type(dwell_solver).__name__ == "DwellSolver"
    assert type(stream_builder).__name__ == "StreamBuilder"
    # make sure the extension is .pickle
    file_path = os.path.splitext(file_path)[0] + '.pickle'
    with open(file_path, 'wb') as f:
        pickle.dump((dwell_solver, stream_builder), f)


def load_build(file_path):
    """Loads the build from the given pickled file.

    Args:
        file_path (str): File path to the pickled file

    Returns:

        dwell_solver (DwellSolver), stream_builder (StreamBuilder)
    """
    with open(file_path, 'rb') as f:
        loaded_data = pickle.load(f)
        dwell_solver, stream_builder = loaded_data[0], loaded_data[1]
    return dwell_solver, stream_builder


def create_safe_savename(path):
    """Creates a path which does not already exist

    Args:
        path (str)

    Returns:
        str: Path same as the given path, but appends a string (e.g. _001) to the path if it already exists
    """
    ext = os.path.splitext(path)[1]
    path_noext = os.path.splitext(path)[0]
    init_path = path_noext
    i = 0
    while len(glob(path_noext + '*')) > 0:
        i += 1
        path_noext = init_path + '_{:03}'.format(i)
    path = path_noext + ext
    return path



import numpy as np
from scipy.optimize import curve_fit
from scipy.spatial import KDTree
from .resistance import get_resistance


class Model:
    """Template class for the model classes. Defines how we model the deposit.
    """

    def __init__(self, struct, get_parameters=True):
        self._struct = struct
        if get_parameters:
            self.get_layer_parameters()

    @property
    def struct(self):
        """Structure to which the model applies.
        """
        return self._struct

    def set_structure(self, struct, update_parameters=True):
        """Sets the structure and updates any internal parameters (e.g. resistance)

        Args:
            struct (Structure): Structure to update.
            update_parameters (bool, optional): Whether or not to update parameters.. Defaults to True.
        """
        self._struct = struct
        if update_parameters:
            self.get_layer_parameters()

    def get_nb_threshold(self):
        """Gets the threshold distance for which the points in a layer are considered neighbours.

        Returns:
            distance (float):
        """
        return 0

    def get_distance_matrix(self, layer):
        """Gets the distance matrix for the layer given by the index.

        Args:
            layer (int,): Index of the layer

        Returns:
            coo_matrix: distance_matrix: Sparse matrix (SciPy coo_matrix) of distances within the points that are withing the nb_threshold as defined by the class
        """
        tree = KDTree(self.struct.slices[layer])
        threshold = self.get_nb_threshold()
        return tree.sparse_distance_matrix(tree, threshold, output_type='coo_matrix')

    def proximity_fun(self, distances, *args):
        """Defines the proximity function to get the proximity matrix from distances.

        Args:
            distances (float, matrix)

        Returns:
            type(distances): proximit value
        """
        return distances + 1

    def get_proximity_matrix(self, layer, *args):
        """Gets the proximity matrix for the layer required by the solver.

        Args:
            layer (int): Index of the layer

        Returns:
            coo_matrix: proximity_matrix: Sparse matrix (SciPy coo_matrix) defining the parameters for the proximity calculation.
        """
        distance_matrix = self.get_distance_matrix(layer)
        # need to copy the matrix so that we can apply the proximity function to the data
        proximity_matrix = distance_matrix.copy()
        proximity_matrix.data = self.proximity_fun(distance_matrix.data, *args)
        return proximity_matrix

    def get_layer_parameters(self):
        """Gets any necessary layer parameters from the structure for the model to be able to calculate the proximity matrix. E.g. resistance for temperature, layer height for focus correction etc."""
        pass


class RRLModel(Model):
    """Reaction rate limited model. Basic model only taking into account growth rate and sigma parameters.

        Attributes:
            gr (float): growth rate
            sigma (float): deposit width
    """

    def __init__(self, struct, gr, sigma, **kwargs):
        super().__init__(struct, **kwargs)
        self.gr = gr
        self.sigma = sigma

    def get_nb_threshold(self):
        """How far are the points considered neighbours"""
        return 3 * self.sigma

    def proximity_fun(self, distances):
        return self.gr * np.exp(-distances**2 / (2 * self.sigma**2))

    @staticmethod
    def calibration_fit_function(t, gr):
        """Function for fitting the calibration"""
        return t * gr

    @staticmethod
    def fit_calibration(dwell_times, lengths, gr0=0.1):
        """Fits the calibration and returns optimal parameters and the fit function."""
        fn = RRLModel.calibration_fit_function

        popt, pcov = curve_fit(fn, dwell_times, lengths,
                               p0=[gr0, ], bounds=(0, np.inf))
        return fn, popt, pcov


class DDModel(Model):
    """Desorption-dominated model taking into account the heating via the resistance model.

        Attributes:
            struct (Structure):
            gr (float): growth rate
            k (float): temperature scaling parameter
            sigma (float): deposit width
    """

    def __init__(self, struct, gr, k, sigma, single_pixel_width=50, **kwargs):
        self.gr = gr
        self.k = k
        self.sigma = sigma

        self.single_pixel_width = single_pixel_width

        self._resistance = None
        super().__init__(struct, **kwargs)

    @property
    def resistance(self):
        """Resistance parameter."""
        if self._resistance is None:
            self.get_layer_parameters()
        return self._resistance

    def get_layer_parameters(self):
        """Gets the resistance and stores it as an internal parameter
        """
        self._resistance = get_resistance(
            self.struct, single_pixel_width=self.single_pixel_width)

    def proximity_fun(self, distances, resistance):
        return self.gr * np.exp(-self.k * resistance) * np.exp(-distances**2 / (2 * self.sigma**2))

    def get_proximity_matrix(self, layer):
        """Returns the proximity matrix using the distance matrix for the given layer."""
        distance_matrix = self.get_distance_matrix(layer)
        # each row of distance matrix has the resistance of the corresponding point
        res = self.resistance[layer][distance_matrix.row]
        # get the proximity matrix
        proximity_matrix = distance_matrix.copy()
        proximity_matrix.data = self.proximity_fun(
            distance_matrix.data, res)
        return proximity_matrix

    def get_nb_threshold(self):
        """How far are the points considered neighbours."""
        return 3 * self.sigma

    @staticmethod
    def calibration_fit_function(t, gr, k):
        """Function for fitting the calibration."""
        return 1 / k * np.log(k * gr * t + 1)

    @staticmethod
    def fit_calibration(dwell_times, lengths, gr0=0.1, k0=1):
        """Fits the calibration and returns optimal parameters and the fit function.

        Args:
            dwell_times (array): Array of measured dwell times.
            lengths (array): Array of measured lengths.
            gr0 (float, optional): Initial guess for GR. Defaults to 0.1.
            k0 (float, optional): Initial guess for k. Defaults to 1.

        Returns:
            [type]: [description]
        """
        fn = DDModel.calibration_fit_function

        popt, pcov = curve_fit(fn, dwell_times, lengths,
                               p0=[gr0, k0], bounds=(0, np.inf))
        print('GR: ', popt[0])
        print('k: ', popt[1])
        return fn, popt, pcov



from numba import njit, config, prange
import numpy.linalg as la
import numpy as np

# set the threading layer before any parallel target compilation
config.THREADING_LAYER = 'threadsafe'


@njit(parallel=True, fastmath=True)
def _numba_eqd_pts(start_nodes, conn_vecs, n_steps):
    pts = np.zeros((np.sum(n_steps), 2))
    end_indices = np.cumsum(n_steps)
    start_indices = np.zeros(end_indices.shape[0]).astype(np.uint32)
    start_indices[1:] = end_indices[:-1]
    for i in prange(n_steps.shape[0]):
        end_indx = end_indices[i]
        start_indx = start_indices[i]
        n = n_steps[i]
        pts[start_indx:end_indx, :] = start_nodes[i, :].reshape(
            1, -1) + conn_vecs[i, :].reshape(1, -1) * np.arange(n).reshape(-1, 1) / n
    return pts


def get_line_eqd_pts(lines, pitch):
    """Gets the line points that are on a grid with spacing pitch.

    Args:
        lines ((n,2,2) array): Array of lines (start_node - end_node)
        pitch (float): Distance on the grid.

    Returns:
        (n,2) array: Points out on the grid.
    """

    start_nodes = np.ascontiguousarray(lines[:, 0, :])
    end_nodes = np.ascontiguousarray(lines[:, 1, :])

    conn_vecs = end_nodes - start_nodes
    conn_norms = la.norm(conn_vecs, axis=1)
    n_steps = np.ceil(conn_norms / pitch).astype(np.int32)

    pts = _numba_eqd_pts(start_nodes, conn_vecs, n_steps)

    pts_unique = np.unique(np.round(pts / pitch) * pitch, axis=0)

    return pts_unique


# def get_lines_length(lines):
#     return np.sum(la.norm(lines[:, 1, :] - lines[:, 0, :], axis=1))

def get_path_length(pts):
    """Gets the length of path defined by moving through the sequence of points

        Args:
            pts ((n,m) array): A sequence of n m-dimensional points

        Returns:
            (n,2) array: Distance of the path moving between the points
    """
    return np.sum(la.norm(pts[1:, :] - pts[:-1, :], axis=1))


def split_eqd(branch_intersections_slices, pitch):
    """Split into equidistant points, keep track of connected components (branches)
    can parallelize with joblib.

    Args:
        branch_intersections_slices (list of lists of arrays): For each slice, for each branch, array of points in that branch.
        pitch (float): Required spacing between the points.

    Returns:
        tuple: slices, branches, branch_lengths
    """
    slices = []
    branches = []
    branch_lengths = []
    for branch_intersections in branch_intersections_slices:
        branches_pts = [get_line_eqd_pts(
            lines, pitch) for lines in branch_intersections]

        slices.append(np.vstack(branches_pts))
        branches.append(np.concatenate(
            [i * np.ones(brpts.shape[0]) for i, brpts in enumerate(branches_pts)]))
        # defining the branch length can be a bit tricky. Here I use the total distance along a path in branches_pts, but this might not be absolutely correct for all structures
        branch_lengths.append(
            [get_path_length(pts) for pts in branches_pts])
    return slices, branches, branch_lengths



from setuptools import find_packages, setup
import os


# load README.md as long_description
long_description = ''
if os.path.exists('README.md'):
    with open('README.md', 'r') as f:
        long_description = f.read()

setup(
    name='f3ast',
    version='1.0.1',
    packages=['f3ast', 'f3ast.calibration'],
    description='FEBID 3D Algorithm for Stream File Generation (F3AST)',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Luka Skoric',
    license='GNU GENERAL PUBLIC LICENSE',
    install_requires=[
        'trimesh>=3.9.12',
        'numpy>=1.20.2',
        'matplotlib>=3.4.1',
        'hjson>=3.0.2',
        'numba>=0.53.1',
        'joblib>=1.0.1',
        'PyQt5>=5.15.4',
        'scikit-image>=0.18.1',
        'scipy>=1.6.2',
        'tbb>=2021.2.0',
        'pyglet>=1.5.16',
        'pytest>=7.1.3',
    ]
)



import numpy as np
from skimage import io
from skimage.color import rgb2gray
from skimage.filters import gaussian, threshold_minimum
from skimage.measure import label
from skimage.color import label2rgb


def read_image(file_path):
    """Reads the image.

    Args:
        file_path (str): File path to the image

    Returns:
        array: Grayscaled image.
    """
    return rgb2gray(io.imread(file_path))


def remove_bottom_bar(img, bottom_factor=0.07):
    """Cuts off the bottom of the image

    Args:
        img (array): Image in array.
        bottom_factor (float, optional): What percentage to cut off. Defaults to 0.07.

    Returns:
        array: Cut image.
    """
    return img[:int((1 - bottom_factor) * img.shape[0]), :]


def threshold_image(img, thresh=None, sigma=2):
    """Smooth and then threshold the image. If thresh is None, try to find the threshold using the Minimum method from skimage.

    Args:
        img (array): Image
        thresh (float, optional): Threshold for the image. If None, try to find it automatically. Defaults to None.
        sigma (float, optional): Blur to apply. Defaults to 2.

    Returns:
        array: Thresholded image.
    """
    img_smooth = gaussian(img, sigma)
    if thresh is None:
        thresh = threshold_minimum(img_smooth)
    img_thresh = img_smooth > thresh
    return img_thresh


def get_labelled_image(img_thresh):
    """Takes the thresholded image and labels the closed areas.

    Args:
        img_thresh (array): Thresholded image.

    Returns:
        tuple: labelled image, label overlay for seeing the labels
    """
    # label image regions
    label_image = label(img_thresh)
    # get the representation of the labellng in overlay image
    image_label_overlay = label2rgb(label_image, image=img_thresh, bg_label=0)
    return label_image, image_label_overlay


def filter_small_labels(label_image, min_struct_size=300):
    """Removes the labels which are of very small areas.

    Args:
        label_image (array): labelled image
        min_struct_size (float, optional): Minimum size. Defaults to 300.

    Returns:
        list: List of valid labels.
    """
    labels = np.unique(label_image.flatten())
    labels = labels[labels != 0]

    # get rid of small labels
    min_struct_size = 300
    to_remove = []
    for l in labels:
        label_size = np.nonzero(label_image.flatten() == l)[0].size
        if label_size < min_struct_size:
            to_remove.append(l)
            label_image[label_image == l] = 0
    labels = [lbl for lbl in labels if lbl not in to_remove]
    return labels


def get_lengths_px(label_image, min_struct_size=300):
    """Gets lengths of labelled structures

    Args:
        label_image (array): labelled image
        min_struct_size (float, optional): Minimum size. Defaults to 300.

    Returns:
        array: Lengths of the valid labelled structures.
    """
    labels = filter_small_labels(label_image, min_struct_size=min_struct_size)
    lengths_px = np.zeros(len(labels))
    for i, lbl in enumerate(labels):
        y_nonzero = np.nonzero(label_image == lbl)[0]
        y_range = np.max(y_nonzero) - np.min(y_nonzero)
        lengths_px[i] = y_range
    return lengths_px



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation

plt.rcParams['image.cmap'] = 'gray'
plt.ion()


class ScaleDisplay(object):
    """Class for selecting a scale. Last two clicks are stored in scale_markers.

        Attributes:
            fig:
            ax:
            img:
            axim:
            scale_markers:
    """

    def __init__(self, fig, ax, img):
        self.fig = fig
        self.ax = ax

        self.img = img

        self.axim = ax.imshow(self.img, cmap='gray')
        # create connections
        self.fig.canvas.mpl_connect('button_press_event', self.onclick)

        # recording of clicks
        self.scale_markers = []

    def onclick(self, event):
        """Prints and saves the click location.
        """
        x = event.xdata
        y = event.ydata
        print('{}, {}'.format(int(np.round(x)), int(np.round(y))))
        if len(self.scale_markers) < 2:
            self.scale_markers.append(x)
        else:
            print('Cleared last selection.')
            self.scale_markers = [x]


def select_scale(img, scale_boundaries=None):
    """Creates axes and plots the image for selecting the scale"""
    fig, ax = plt.subplots(1, 1, figsize=[10, 4])
    if scale_boundaries is None:
        scale_boundaries = [int(img.shape[0] * 0.93), int(img.shape[1] * 0.6)]
    img1 = img[scale_boundaries[0]:, scale_boundaries[1]:]
    tracker = ScaleDisplay(fig, ax, img1)
    return tracker



from scipy.optimize import lsq_linear
import numpy as np
from .plotting import plot_dwells
from scipy.spatial import KDTree
from datetime import timedelta
import time
from joblib import Parallel, delayed


def get_distance_matrix(sl, threshold):
    """Gets the sparse matrix containting distances between points i and j in slice sl that are under a threshold.

    Args:
        sl ((n,2) array): Points in slice to which to get the distance matrix of.
        threshold (float): Maximal distance which to consider.

    Returns:
        coo_matrix: Sparse distance matrix.
    """
    tree = KDTree(sl)
    return tree.sparse_distance_matrix(tree, threshold, output_type='coo_matrix')


class DwellSolver:
    """Class which solves the proximity problem for dwell times.

        Attributes:
            model (Model): Model which to solve.
            dwell_times_slices (list of arrays): Solutionto the proximity problem.
    """

    def __init__(self, model):
        self.model = model
        self.dwell_times_slices = None

    def solve_dwells(self, n_jobs=5):
        """Solves the dwells for dwell times and stores the result in self.dwell_times_slices

        Args:
            tol (float, optional): Tolerance of the lsq_linear solver. Defaults to 1e-3.
        """
        print('Solving for dwells...')
        # get the thickness of layers
        dz_slices = self.model.struct.dz_slices
        n_layers = dz_slices.size
        # get the generator for the proximity matrix
        prox_matrix_generator = (self.model.get_proximity_matrix(lyr)
                                 for lyr in range(n_layers))
        dwell_times_slices = list()
        # solve for each layer. Do this in parallel to speed up.
        dwell_times_slices = Parallel(n_jobs=n_jobs)(delayed(self.solve_layer)(
            proximity_matrix, dz) for proximity_matrix, dz in zip(prox_matrix_generator, dz_slices))
        self.dwell_times_slices = dwell_times_slices
        print('Solved')

    @staticmethod
    def solve_layer(proximity_matrix, dz, tol=1e-3):
        """Solves a layer proximity problem given a proximity matrix and the layer height.

        Args:
            proximity_matrix (sparse matrix): Proximity matrix
            dz (float): Layer height
            tol (float, optional): Tolerance for the optimization. Defaults to 1e-3.

        Returns:
            dwell_times: Array of dwell times as a solution for the layer.
        """
        # get a tight upper bound for faster computation. We can never have larger dwell times than if there was no proximity (proximity matrix was diagonal)
        upper_bound = dz / proximity_matrix.diagonal()
        # solve the optimization problem
        y = dz * np.ones(proximity_matrix.shape[1])
        result = lsq_linear(proximity_matrix, y,
                            bounds=(0, upper_bound), tol=tol)
        return result.x

    def get_dwells_slices(self):
        """Returns the dwells for point as a per-slice list"""
        if self.dwell_times_slices is None:
            self.solve_dwells()
        slices3d = self.model.struct.get_3dslices()
        return [np.hstack([dwt[:, np.newaxis], sl3d]) for dwt, sl3d in zip(
                self.dwell_times_slices, slices3d)]

    def get_dwells_matrix(self):
        """Returns the concatenated array of dwells for each point in a Nx4 array (time, x, y, z)"""
        return np.vstack(self.get_dwells_slices())

    def show_solution(self, cutoff=None):
        """Plots the solution colouring by dwells and displaying only the dwells above the cutoff.

        Args:
            cutoff (float, optional): Minimum dwell time to include. Defaults to None.

        Returns:
            tuple: axes, sc
        """
        dwells = self.get_dwells_matrix()
        if cutoff is not None:
            dwells = dwells[dwells[:, 0] > cutoff, :]
        ax, sc = plot_dwells(dwells)
        return ax, sc

    def get_total_time(self):
        """Gets the total solution time. Returns None if not solved.

        Returns:
            datetime.timedelta:
        """
        if self.dwell_times_slices is not None:
            t = np.sum([np.sum(dwls) for dwls in self.dwell_times_slices])
            return timedelta(milliseconds=t)
        return None

    def print_total_time(self):
        """Prints the total stream time"""
        t = self.get_total_time()
        if t is not None:
            print('Total stream time: ', t)
        else:
            print('Dwell times not calculated yet!')
